import * as THREE from 'three';
import { GameOptions } from './gameOptions';
import './style.css';

const xLimit : number = GameOptions.trackSideLanes * GameOptions.tileSize + GameOptions.tileSize / 2;                               // x limit, to keep the ball inside the track
const zLimit : number = -(GameOptions.trackLenght - 1) * GameOptions.tileSize;  // final z position to reach to close the track

let tiles       : THREE.Mesh[]  = [];
let speed       : number        = GameOptions.startSpeed;   // player speed
let velocityX   : number        = 0;                        // x velocity (to move left and right)                        
let moveLeft    : boolean       = false;                    // are we moving left?
let moveRight   : boolean       = false;                    // are we moving right?
let speedUp     : boolean       = false;                    // are we accelerating?
let slowDown    : boolean       = false;                    // are we braking?
let gameStopped : boolean       = false;                    // is the game stopped?
let lastTime    : number        = performance.now();        // current timestamp

// create a new Three scene
const scene : THREE.Scene = new THREE.Scene();

// add a black fog to the scene
scene.fog = new THREE.Fog(0x000000, 10, 40);

// add a camera
const camera : THREE.PerspectiveCamera = new THREE.PerspectiveCamera(
    75,                                     // field of view
    window.innerWidth / window.innerHeight, // aspect ratio
    0.1,                                    // near rendering
    1000                                    // far rendering
);
camera.position.set(0, GameOptions.cameraZDistance / 3 * 2, GameOptions.cameraZDistance);

// add a renderer
const renderer : THREE.WebGLRenderer = new THREE.WebGLRenderer({
    antialias : true
});

// set renderer size
renderer.setSize(window.innerWidth, window.innerHeight);

// add renderer to the page
document.body.appendChild(renderer.domElement);

// add a default ambient light
const ambientLight : THREE.AmbientLight = new THREE.AmbientLight();
scene.add(ambientLight);

// add ball and shadow
const ball : THREE.Mesh = addBall();
const shadow : THREE.Mesh = addShadow();

// generate the track
generateTrack();

// update the game
update();

// function to add the ball
function addBall() : THREE.Mesh {

    // load texture for the ball
    const textureLoader : THREE.TextureLoader = new THREE.TextureLoader();
    const ballTexture : THREE.Texture = textureLoader.load('assets/sprites/ball.png');

    // create ball material
    const ballMaterial : THREE.MeshStandardMaterial = new THREE.MeshStandardMaterial({ 
        map : ballTexture
    });

    // create ball geometry
    const ballGeometry : THREE.SphereGeometry = new THREE.SphereGeometry(GameOptions.ballRadius);

    // create the ball itself
    const ball : THREE.Mesh = new THREE.Mesh(ballGeometry, ballMaterial);

    // set ball position
    ball.position.set(0, GameOptions.ballRadius, 0);

    // add the ball to the scene
    scene.add(ball);

    return ball;
}

// function to add the shadow
function addShadow() : THREE.Mesh {

    // add thexture for the shadow
    const textureLoader : THREE.TextureLoader = new THREE.TextureLoader();
    const shadowTexture : THREE.Texture = textureLoader.load('assets/sprites/shadow.png');
    
    // create shadow material
    const shadowMaterial : THREE.MeshStandardMaterial = new THREE.MeshStandardMaterial({
        map         : shadowTexture,
        transparent : true, 
    });

    // create shadow geometry
    const shadowGeometry = new THREE.PlaneGeometry(GameOptions.ballRadius * 2, GameOptions.ballRadius * 2);

    // create the shadow itself
    const shadow = new THREE.Mesh(shadowGeometry, shadowMaterial);

    // place the shadow horizontally
    shadow.rotation.x = -Math.PI / 2; // piano orizzontale

    // place it a bit over tiles
    shadow.position.y = 0.06; 

    // add the shadow to the scene
    scene.add(shadow);

    return shadow;
}

// function to generate the track
function generateTrack() : void {
    
    // remove all existing tiles
    for (const tile of tiles) {
        scene.remove(tile);
    }

    // rest tiles array
    tiles = [];

    // build the track, tile by tile
    for (let i : number = 0; i < GameOptions.trackLenght; i ++) {
        for (let j : number = - GameOptions.trackSideLanes; j <= GameOptions.trackSideLanes; j ++) {
            
            // pick a random color
            let color : number = GameOptions.tileColors[Math.floor(Math.random() * GameOptions.tileColors.length)];

            // if it's the finish line, add finish line colors
            if (i == GameOptions.trackLenght - 1) {
                console.log(j % 2)
                color = GameOptions.finishColors[Math.abs(j) % 2];
            }
            
            // define tile geometry
            const tileGeometry : THREE.BoxGeometry = new THREE.BoxGeometry(GameOptions.tileSize, 0.1, GameOptions.tileSize);

            // define tile material
            const tileMaterial : THREE.MeshStandardMaterial = new THREE.MeshStandardMaterial({
                color : color
            });

            // create tile mesh
            const tile: THREE.Mesh = new THREE.Mesh(tileGeometry, tileMaterial);
           
            // set tile position
            tile.position.set(j * GameOptions.tileSize, 0, -i * GameOptions.tileSize);

            // add the tile to the scene
            scene.add(tile);

            // add the tile to tiles array
            tiles.push(tile);
        }
    }
}

// function to reset the game
function resetGame() : void {

    // generate a new track
    generateTrack();

    // place the ball in its position
    ball.position.set(0, GameOptions.ballRadius, 0);
    ball.rotation.set(0, 0, 0);

    // reset the speed
    speed = GameOptions.startSpeed;
    velocityX = 0;

    // unstop the game
    gameStopped = false;
}

// function to be called at each frame
function update() : void {

    // request animation frame
    requestAnimationFrame(update);

    // get the amount of seconds since previous frame
    const currentTime : number = performance.now();
    const deltaTime : number = (currentTime - lastTime) / 1000; 
    lastTime = currentTime;

    // if the game is stopped, that's all
    if (gameStopped) {
        return;
    }

    // handle acceleration
    if (speedUp) {
        speed += GameOptions.acceleration * deltaTime;
    }

    // handle brakes, with a minimum speed
    if (slowDown) {
        speed = Math.max(GameOptions.minSpeed, speed - GameOptions.brake * deltaTime);
    }

    // handle left steering
    if (moveLeft && !moveRight) {
        velocityX = Math.max(velocityX - GameOptions.sideAcceleration * deltaTime, -GameOptions.maxXVelocity);
    }
    else {

        // handle right steering
        if (moveRight && !moveLeft) {
            velocityX = Math.min(velocityX + GameOptions.sideAcceleration * deltaTime, GameOptions.maxXVelocity);
        }   
        else {

            // no steering
            velocityX *= GameOptions.xDeceleration;    
        } 
    }
   
    // update ball position
    ball.position.z -= speed;
    ball.position.x += velocityX;

    // keep the ball inside the track
    if (ball.position.x > xLimit) {
        ball.position.x = xLimit;
        velocityX = 0;
    }
    if (ball.position.x < -xLimit) {
        ball.position.x = -xLimit;
        velocityX = 0;
    }

    // rotate ball according to speed
    ball.rotation.x -= speed / GameOptions.ballRadius;

    // update shadow position
    shadow.position.x = ball.position.x;
    shadow.position.z = ball.position.z;

    // update camera to follow the ball
    camera.position.z = ball.position.z + GameOptions.cameraZDistance;
    camera.lookAt(new THREE.Vector3(0, ball.position.y, ball.position.z));

    // if we reached the finish line, stop and restart
    if (ball.position.z <= zLimit) {
        speed = 0;
        velocityX = 0;
        gameStopped = true;
        setTimeout(resetGame, 2000);
    }

    // render the scene
    renderer.render(scene, camera);
}

// key down listener, to trigger WASD movements
window.addEventListener('keydown', (e: KeyboardEvent) : void => {
    switch (e.key.toLowerCase()) {
      case 'a':
        moveLeft = true;
        break;
      case 'd':
        moveRight = true;
        break;
      case 'w':
        speedUp = true;
        break;
      case 's':
        slowDown = true;
        break;
    }
});

// key up listener
window.addEventListener('keyup', (e: KeyboardEvent) : void => {
    switch (e.key.toLowerCase()) {
      case 'a':
        moveLeft = false;
        break;
      case 'd':
        moveRight = false;
        break;
      case 'w':
        speedUp = false;
        break;
      case 's':
        slowDown = false;
        break;
    }
});

// resize window listener
window.addEventListener('resize', () : void => {
    camera.aspect = window.innerWidth / window.innerHeight;
    camera.updateProjectionMatrix();
    renderer.setSize(window.innerWidth, window.innerHeight);
});
